#!/usr/bin/env bash
# Copyright (c) 2004-2023 Yury V. Vishnevskiy yu.v.vishnevskiy@gmail.com
# All Rights Reserved.
# This file is subject to the terms and conditions defined in the
# 'license.txt' file, which is part of this program package.

sargs=("$@")

set -fb

function usage() {
    if [ -n "$1" ]; then
        echo -e "$1\n";
    fi
    echo "Usage: $0 [-y,--assume-yes] [-h,--help]"
    echo "  -y,--assume-yes  Automatic yes to prompts"
    echo "  -h,--help        Show help message"
    echo ""
    exit 1
}


# parse params
for ((i=0; i < ${#sargs[@]}; i++))
do
    item=${sargs[i]}
    case "${item}" in
        "-y" | "--assume-yes")
            FORCE_YES=1
            ;;
	"-h" | "--help")
	    usage
	    ;;
        *)
            usage "Unknown parameter passed: $1"
            ;;
    esac
done


if [[ -t 1 ]]; then
    # stdout is a terminal
    GREEN='\033[0;32m'
    RED='\033[0;31m'
    NC='\033[0m' # No Color
else
    # stdout is not a terminal
    GREEN=''
    RED=''
    NC=''
fi


readonly UDIR=$(cd "$(dirname "$0")" || exit 1 ; pwd)
readonly MYNAME=$(basename "$0")
readonly EXISTING_SCRIPT="${UDIR}/update.sh"
readonly EXECUTABLE_SCRIPT="${UDIR}/.update.sh"

UURLS="https://unex.vishnevskiy.group/files/"
UURL=""

TMPDIR=""

UPDRDY=False

function clean_up() {
  # Here is the clean up code that has to execute every time
  echo "Deleting temporary files."
  [ -z "$TMPDIR" ] || rm -rf "${TMPDIR}"
}


function self_clean_up() {
  echo "Self clean up."
  rm -f "${EXECUTABLE_SCRIPT}"
}


function copy_self_and_invoke() {
  cp "${EXISTING_SCRIPT}" "${EXECUTABLE_SCRIPT}"
  exec "${EXECUTABLE_SCRIPT}" "${sargs[@]}"
}


function get_filehash() {
    # $1 path to file
    retc=0
    if [ "$(command -v sha256sum)" ]; then
        sha256str=$(sha256sum "$1" 2>&1)
        if [ $? -ne 0 ]; then
            retstr=$sha256str
            retc=1
        else
            retstr=$(echo "$sha256str" | awk '{print $1}')
        fi
    elif [ "$(command -v shasum)" ]; then
        sha256str=$(shasum -a 256 "$1" 2>&1)
        if [ $? -ne 0 ]; then
            retstr=$sha256str
            retc=1
        else
            retstr=$(echo "$sha256str" | awk '{print $1}')
        fi
    else
        retstr="Cannot find hash utility in PATH."
        retc=1
    fi
    echo "$retstr"
    return $retc
}


function download_file() {
    # $1 URL
    # $2 path to file
    retc=0
    if [ "$(command -v wget)" ]; then
        retstr=$(wget -O "$2" "$1" 2>&1)
        if [ $? -ne 0 ]; then
            retc=1
        fi
    elif [ "$(command -v curl)" ]; then
        retstr=$(curl "$1" -o "$2" 2>&1)
        if [ $? -ne 0 ]; then
            retc=1
        fi
    else
        retstr="Cannot find wget or curl utilities in PATH."
        retc=1
    fi
    echo "$retstr"
    return $retc
}


function update_license() {
    lhtmlicf="${UDIR}/license.html"
    ltxtlicf="${UDIR}/license.txt"
    lpdflicf="${UDIR}/license.pdf"
    rhtmlicf="${TMPDIR}/license.html"
    htmlic_file_url="${UURL}license.html"
    txtlic_file_url="${UURL}license.txt"
    pdflic_file_url="${UURL}license.pdf"
    downstr=$(download_file $htmlic_file_url $rhtmlicf)
    if [ $? -ne 0 ]; then echo "$downstr"; return 1; fi
    rlichash=$(get_filehash $rhtmlicf)
    if [ $? -ne 0 ]; then echo "$rlichash"; exit 1; fi
    if [ -f "$lhtmlicf" ]; then
	llichash=$(get_filehash "$lhtmlicf")
	if [ $? -ne 0 ]; then echo "$llichash"; exit 1; fi
    else
	llichash="qwerty"
    fi
    if [ "$rlichash" != "$llichash" ]; then
	#Backup old files
	for licf in $lhtmlicf $lpdflicf $ltxtlicf; do
	    if [ -f "$licf" ]; then
		dirn=$(dirname "$licf")
		basen=$(basename "$licf")
		extn="${basen##*.}"
		filen="${basen%.*}"
		renamed="${dirn}/${filen}_old.${extn}"
		cp "$licf" "$renamed"
	    fi
	done
	cp "$rhtmlicf" "$lhtmlicf"
	if [ $? -ne 0 ]; then printf "${RED}Cannot copy new license file!${NC}\n"; return 1; fi
	downstr=$(download_file $pdflic_file_url "$lpdflicf")
	if [ $? -ne 0 ]; then rm -f "$lpdflicf"; fi
	downstr=$(download_file $txtlic_file_url "$ltxtlicf")
	if [ $? -ne 0 ]; then rm -f "$ltxtlicf"; fi
	printf "${GREEN}Warning, license has been updated!${NC}\n"
    fi
}


UPROGSTR=""
UVERSTR=""
UMAJORVER=""
UMINORVER=""
UCOMMNUM=""
UFILEB=""

# Parsing complete version file
function parse_uverf() {
    # Path to file in $1
    if [ -f "$1" ]; then
	UPROGSTR=$(head -n 1 "$1" | tail -1)
	UVERSTR=$(head -n 2 "$1" | tail -1)
	losstr=$(head -n 3 "$1" | tail -1)
	larchstr=$(head -n 4 "$1" | tail -1)
	lbldstr=$(head -n 5 "$1" | tail -1)
    else
	printf "${RED}$1 does not exist!${NC}\n"
	return 1
    fi

    # UVERSTR format: 2.0-1567-g2f1d8f5b (2019-09-02)
    UMAJORVER=$(echo "$UVERSTR" | cut -f1 -d- | cut -f1 -d.)
    UMINORVER=$(echo "$UVERSTR" | cut -f1 -d- | cut -f2 -d.)
    UCOMMNUM=$(echo "$UVERSTR" | cut -f2 -d-)

    if [ "$losstr" = "Linux" ]; then
	fosstr="linux"
    elif [ "$losstr" = "macOS" ]; then
	fosstr="macos"
    elif [ "$losstr" = "FreeBSD" ]; then
	fosstr="freebsd"
    elif [ "$losstr" = "Windows" ]; then
	fosstr="win"
    else
	printf "${RED}$losstr is not a valid OS type!${NC}\n"
	return 1
    fi

    if [ "$larchstr" = "x86_64/Core2" ]; then
	farchstr="x86-64-core2"
    elif [ "$larchstr" = "x86/i686" ]; then
	farchstr="x86-i686"
    else
	printf "${RED}$larchstr is not a valid arch type!${NC}\n"
	return 1
    fi

    if [ "$lbldstr" = "Release" ]; then
	fbldstr="release"
    elif [ "$lbldstr" = "Debug" ]; then
	fbldstr="debug"
    else
	printf "${RED}$lbldstr is not a valid build type!${NC}\n"
	return 1
    fi

    proglcs=$(echo "$UPROGSTR" | tr '[:upper:]' '[:lower:]')
    UFILEB="${proglcs}_${fosstr}_${farchstr}_${fbldstr}"
}


# Parsing short version file
function parse_usverf() {
    # Path to file in $1
    # UVERSTR format: 1.6-1430 (2023-08-07)
    if [ -f "$1" ]; then
	UVERSTR=$(head -n 1 "$1" | tail -1)
    else
	printf "${RED}$1 does not exist!${NC}\n"
	return 1
    fi

    UCOMMNUM=$(echo "$UVERSTR" | cut -f1 -d' ' | cut -f2 -d-)
    UMAJORVER=$(echo "$UVERSTR" | cut -f1 -d' ' | cut -f1 -d- | cut -f1 -d.)
    UMINORVER=$(echo "$UVERSTR" | cut -f1 -d' ' | cut -f1 -d- | cut -f2 -d.)
}


function validate_url() {
    # $1 URL
    retc=1
    if [ "$(command -v wget)" ]; then
        if [[ $(wget --max-redirect=0 -S --spider "$1" 2>&1 | grep 'HTTP/1.1 200 OK') ]]; then
            retc=0
        fi
    elif [ "$(command -v curl)" ]; then
        if [[ $(curl -Is "$1" 2>&1 | grep 'HTTP/1.1 200 OK') ]]; then
            retc=0
        fi
    else
        echo "Cannot find wget or curl utilities in PATH."
    fi
    return $retc
}


function find_best_url() {
    luverf="${UDIR}/version.txt"
    parse_uverf "$luverf"
    if [ $? -ne 0 ]; then printf "${RED}Cannot parse local version file!${NC}\n"; exit 1; fi
    lverstr=$UVERSTR
    lmajorver=$UMAJORVER
    lminorver=$UMINORVER
    lcommnum=$UCOMMNUM
    ufileb=$UFILEB
    ruverf="${TMPDIR}/${ufileb}.txt"
    echo "Local version: ${lverstr}"

    for i in $UURLS ; do
        uver_file_url="${i}${ufileb}.txt"
        if ! validate_url "$uver_file_url"; then
            continue
        else
            UURL=$i

            downstr=$(download_file "$uver_file_url" "$ruverf")
            if [ $? -ne 0 ]; then
                printf "${RED}Cannot download ${uver_file_url}${NC}\n"
                echo "$downstr"
                exit 1
            fi

            parse_usverf "$ruverf"
            if [ $? -ne 0 ]; then continue; fi
            rverstr=$UVERSTR
            rmajorver=$UMAJORVER
            rminorver=$UMINORVER
            rcommnum=$UCOMMNUM

            # Check priorities: majorver > minorver > commnum
            if [ "$rmajorver" -gt "$lmajorver" ]; then
                UPDRDY=True
            elif [ "$rmajorver" -eq "$lmajorver" ]; then
                if [ "$rminorver" -gt "$lminorver" ]; then
                    UPDRDY=True
                elif [ "$rminorver" -eq "$lminorver" ]; then
                    if [ "$rcommnum" -gt "$lcommnum" ]; then
                        UPDRDY=True
                    fi
                fi
            fi

            if [ $UPDRDY = True ]; then
                printf "${GREEN}An update for $UPROGSTR is available: ${rverstr}${NC}\n"
                break
            fi
        fi
    done
}


function update_program() {
    ufileb=$UFILEB

    # Try tar.gz first
    uarc_ext="tar.gz"
    uarc_file_url="${UURL}${ufileb}.${uarc_ext}"
    if ! validate_url "$uarc_file_url"; then
        uarc_ext="zip"
        uarc_file_url="${UURL}${ufileb}.${uarc_ext}"
    fi

    echo "Downloading file $uarc_file_url"
    echo "This may take some time."
    rarcf="${TMPDIR}/${ufileb}.${uarc_ext}"
    downstr=$(download_file "$uarc_file_url" "$rarcf")
    if [ $? -ne 0 ]; then
        printf "${RED}Cannot download ${uarc_file_url}${NC}\n"
        echo "$downstr"
        exit 1
    fi

    echo "Unpacking archive."
    if [ "$uarc_ext" = "tar.gz" ]; then
        tar -xzf "${rarcf}" -C "${TMPDIR}" &> "${TMPDIR}/unpack.log"
        rc=$?
    elif [ "$uarc_ext" = "zip" ]; then
        unzip -d "${TMPDIR}" "${rarcf}" &> "${TMPDIR}/unpack.log"
        rc=$?
    else
         printf "${RED}Invalid archive type!${NC}\n"
         exit 1
    fi

    if [ $rc -ne 0 ]; then
         printf "${RED}Cannot unpack ${rarcf}${NC}\n"
         echo "Unpacking log:"
         cat "${TMPDIR}/unpack.log"
         exit 1
    fi

    echo "Warning!"
    echo "Old files in the $UPROGSTR directory will be deleted now."
    echo "$UPROGSTR directory: ${UDIR}"
    if [ -z "$FORCE_YES" ]; then
        while true; do
            IFS= read -rp "Do you want to proceed? (Y/n)" yn
            case $yn in
                "y" | "Y" | "") break;;
                *) echo "Update terminated."; exit 1;;
            esac
        done
    fi

    # First remove old files
    echo "Deleting old files."
    curDir=$(pwd)
    cd "${UDIR}" || exit 1
    ofiles=$(cat files.txt)
    while IFS= read -r line; do
        rm -f "$line"
    done <<< "$ofiles"

    # Copy all files from the downloaded archive
    cd "$curDir" || exit 1
    echo "Copying new files."
    cp -r "${TMPDIR}/${ufileb}/." "${UDIR}"
    if [ $? -eq 0 ]; then
        printf "${GREEN}Updating done.${NC}\n"
    else
        printf "${RED}Cannot copy (some of) new files!${NC}\n"
        exit 1
    fi

    # Delete empty dirs
    echo "Deleting empty directories."
    find "${UDIR}" -type d -empty -delete
}


# Main function
# Updating files is done here
function main() {
    TMPDIR=$(mktemp -d)
    if [ $? -ne 0 ]; then printf "${RED}Cannot create temporary directory!${NC}\n"; exit 1; fi

    find_best_url

    if [ -n "$UURL" ]; then
        update_license
    fi

    if [ $UPDRDY = True ]; then
        update_program
    elif [ -n "$UURL" ]; then
        echo "Local version is up to date."
    else
        printf "${RED}Cannot locate server with updates!${NC}\n"
        printf "${RED}Try to update manually!${NC}\n"
        exit 1
    fi
}

if [[ $MYNAME = \.* ]]; then
  # Copied script
  trap "clean_up; self_clean_up" EXIT
  main
else
  # Original scipt
  trap clean_up EXIT
  copy_self_and_invoke
fi
